#define _GNU_SOURCE

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/fcntl.h>

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>

#include "log.h"
#include "file_cache.h"
#include "fdevent.h"
#include "etag.h"

#ifdef HAVE_ATTR_ATTRIBUTES_H
#include <attr/attributes.h>
#endif

#ifndef O_LARGEFILE
# define O_LARGEFILE 0
#endif

#undef USE_LINUX_SIGIO

file_cache *file_cache_init(void) {
	file_cache *fc = NULL;
	
	fc = calloc(1, sizeof(*fc));
	
	return fc;
}

static file_cache_entry * file_cache_entry_init(void) {
	file_cache_entry *fce = NULL;
	
	fce = calloc(1, sizeof(*fce));
	
	fce->fd = -1;
	fce->fde_ndx = -1;
	fce->name = buffer_init();
	fce->etag = buffer_init();
	fce->content_type = buffer_init();
	
	return fce;
}

static void file_cache_entry_free(file_cache_entry *fce) {
	if (!fce) return;
	
	if (fce->fd >= 0) close(fce->fd);
	
	buffer_free(fce->etag);
	buffer_free(fce->name);
	buffer_free(fce->content_type);
	
	if (fce->mmap_p) munmap(fce->mmap_p, fce->mmap_length);
	
	free(fce);
}

static int file_cache_entry_reset(file_cache_entry *fce) {
	if (fce->fd < 0) return 0;
	
	close(fce->fd);
			
#ifdef USE_LINUX_SIGIO
	/* doesn't work anymore */
	if (fce->fde_ndx != -1) {
		fce->fde_ndx = fdevent_event_del(srv->ev, fce->fde_ndx, fce->fd);
		fdevent_fdc_del(srv->ev->fdc, fce->fd);
	}
#endif
	
	if (fce->mmap_p) {
		munmap(fce->mmap_p, fce->mmap_length); 
		fce->mmap_p = NULL;
	}
	fce->fd = -1;
	
	return 0;
}

void file_cache_free(file_cache *fc) {
	size_t i;
	for (i = 0; i < fc->used; i++) {
		file_cache_entry_free(fc->ptr[i]);
	}
	
	free(fc->ptr);
	
	free(fc);
	
}

#ifdef HAVE_XATTR
int fce_attr_get(buffer *buf, char *name) {
	int attrlen;
	int ret;
	attrlen = 1024;
	buffer_prepare_copy(buf, attrlen);
	if(0 == (ret = attr_get(name, "Content-Type", buf->ptr, &attrlen, 0))) {
		buf->used = attrlen + 1;
		buf->ptr[attrlen] = '\0';
	}
	return ret;
}
#endif

file_cache_entry * file_cache_get_entry(server *srv, connection *con, buffer *name, file_cache_entry *o_fce) {
	file_cache_entry *fce = NULL;
	file_cache *fc = srv->file_cache;
	size_t i;
	size_t oldest_ndx;
	
	UNUSED(con);

	/* still valid ? */
	if (o_fce != NULL) {
		if (buffer_is_equal(name, o_fce->name) &&
		    (o_fce->fd != -1) &&
		    (o_fce->stat_ts == srv->cur_ts)
		    ) {
			return o_fce;
		} else {
			o_fce->in_use--;
		}
	}
	
	oldest_ndx = fc->used;
	
	/* check the cache */
	for (i = 0; i < fc->used; i++) {
		fce = fc->ptr[i];
		
		if (buffer_is_equal(name, fce->name)) {

#ifdef USE_LINUX_SIGIO
			if (fce->is_dirty == 0) {
				fce->in_use++;
				return fce;
			}
#endif
			
			if (-1 == (stat(name->ptr, &(fce->st)))) {
				fce->in_use = 0;
				
				log_error_write(srv, __FILE__, __LINE__, "sbs", "stat failed: ", name, strerror(errno));
				
				file_cache_entry_reset(fce);
				
				return NULL;
			}
			fce->stat_ts = srv->cur_ts;

#ifdef HAVE_XATTR
			fce_attr_get(fce->content_type, name);
#endif
			
			etag_create(srv->file_cache_etag, &(fce->st));
			
			if (!buffer_is_equal(srv->file_cache_etag, fce->etag)) {
				/* etag has changed: reopen file */
				
				file_cache_entry_reset(fce);
				
				if (-1 == (fce->fd = open(fce->name->ptr, O_RDONLY | O_LARGEFILE))) {
					log_error_write(srv, __FILE__, __LINE__, "ss", "open failed: ", strerror(errno));
					
					buffer_reset(fce->name);
					return NULL;
				}
			}
			
			buffer_copy_string_buffer(fce->etag, srv->file_cache_etag);
			
			
#ifdef USE_LINUX_SIGIO
			fce->is_dirty = 0;
#endif
			
			fce->in_use++;
			return fce;
		}
		
		if (fce->in_use == 0 &&
		    ((oldest_ndx == fc->used) ||
		     (fce->stat_ts < fc->ptr[oldest_ndx]->stat_ts))) {
			oldest_ndx = i;
		}
	}
	
	/* not found, insert */
	if (fc->size == 0) {
		fc->size = 16;
		fc->ptr = calloc(fc->size, sizeof(*fc->ptr));
		fc->used = 0;
	}
	
	if (fc->used < fc->size) {
		fce = file_cache_entry_init();
		fc->ptr[fc->used++] = fce;
	} else if (oldest_ndx != fc->used) {
		/* the place is unused, reuse it */
		
		fce = fc->ptr[oldest_ndx];
		
		file_cache_entry_reset(fce);
	} else {
		/* the cache is full, time to resize */
		
		fc->size += 16;
		fc->ptr = realloc(fc->ptr, sizeof(*fc->ptr) * fc->size);
		
		fce = file_cache_entry_init();
		fc->ptr[fc->used++] = fce;
	}
	
	buffer_copy_string_buffer(fce->name, name);
	fce->in_use = 0;
	fce->fd = -1;
	
	if (-1 == stat(name->ptr, &(fce->st))) {
		int oerrno = errno;
#if 0
		log_error_write(srv, __FILE__, __LINE__, "sbs", "stat failed: ", name, strerror(errno));
#endif
		file_cache_entry_reset(fce);
		
		errno = oerrno;
		buffer_reset(fce->name);
		return NULL;
	}
	
	fce->stat_ts = srv->cur_ts;
	
	if (S_ISREG(fce->st.st_mode)) {
		if (-1 == (fce->fd = open(name->ptr, O_RDONLY | O_LARGEFILE))) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "open failed: ", strerror(errno));
			
			buffer_reset(fce->name);
			return NULL;
		}
		
#ifdef HAVE_XATTR
		fce_attr_get(fce->content_type, name->ptr);
#endif
		
		etag_create(fce->etag, &(fce->st));
	} else if (S_ISDIR(fce->st.st_mode)) {
#ifdef USE_LINUX_SIGIO
		/* register sigio for the directory */
		
		if (-1 == (fce->fd = open(name->ptr, O_RDONLY))) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "open failed: ", strerror(errno));
			buffer_reset(fce->name);
			
			return NULL;
		}
		
		if (fcntl(fce->fd, F_NOTIFY, DN_CREATE|DN_DELETE|DN_MODIFY|DN_MULTISHOT) < 0) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "fcntl failed: ", strerror(errno));
			buffer_reset(fce->name);
			return NULL;
		}
		
		/* register fd-handler */
		fdevent_register(srv->ev, con->fcgi.fd, file_cache_handle_fdevent, fce);
		
		fce->fde_ndx = fdevent_event_add(srv->ev, -1, fce->fd, FDEVENT_IN);
# if 0		
		log_error_write(srv, __FILE__, __LINE__, "sddb", "fdevent_event_add: ", fce->fde_ndx, fce->fd, fce->name);
# endif
#endif
	}
	
	fce->in_use++;
		
	return fce;
}

int file_cache_entry_release(server *srv, connection *con, file_cache_entry *fce) {
	UNUSED(srv);
	UNUSED(con);

	if (fce->in_use > 0) fce->in_use--;
	
	return 0;
}


int file_cache_handle_fdevent(server *srv, file_cache_entry *fce, int revent) {
	size_t i;
	file_cache *fc = srv->file_cache;;

	UNUSED(revent);
	/* */
#if 0	
	log_error_write(srv, __FILE__, __LINE__, "sds", "dir has changed: ", fce->fd, fce->name->ptr);
#endif	
	/* touch all files below this directory */
	
	for (i = 0; i < fc->used; i++) {
		file_cache_entry *f = fc->ptr[i];
		
		if (fce == f) continue;
		
		if (0 == strncmp(fce->name->ptr, f->name->ptr, fce->name->used - 1)) {
#if 0
			log_error_write(srv, __FILE__, __LINE__, "ss", "file hit: ", f->name->ptr);
#endif
			f->is_dirty = 1;
		}
	}
	
	return 0;
}
